VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsCrypt"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit
'================================
'EVENTS
'================================
Public Event Error(ByVal Number As Long, Description As String, ByVal Source As String)

'BCryptGetProperty strings (subset used here).
Private Const MS_PRIMITIVE_PROVIDER As String = "Microsoft Primitive Provider"
Private Const BCRYPT_OBJECT_LENGTH As String = "ObjectLength"
Private Const BCRYPT_BLOCK_LENGTH As String = "BlockLength"
Private Const BCRYPT_CHAINING_MODE As String = "ChainingMode"
Private Const BCRYPT_CHAIN_MODE_GCM As String = "ChainingModeGCM"
Private Const BCRYPT_AUTH_TAG_LENGTH As String = "AuthTagLength"

'Constants for Cryptography API error messages
Private Const CE As String = "BCryptEncrypt"
Private Const CGP As String = "BCryptGetProperty"
Private Const CGSK As String = "BCryptGenerateSymmetricKey"
Private Const COAP As String = "BCryptOpenAlgorithmProvider"

'CNG API Declares
Private Declare Function BCryptOpenAlgorithmProvider Lib "bcrypt" (ByRef hAlgorithm As Long, ByVal pszAlgId As Long, ByVal pszImplementation As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptCloseAlgorithmProvider Lib "bcrypt" (ByVal hAlgorithm As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptDestroyKey Lib "bcrypt" (ByVal hAesKey As Long) As Long
Private Declare Function BCryptGetProperty Lib "bcrypt" (ByVal hObject As Long, ByVal pszProperty As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByRef cbResult As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptSetProperty Lib "bcrypt" (ByVal hObject As Long, ByVal pszProperty As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptGenerateSymmetricKey Lib "bcrypt" (ByVal hAlgorithm As Long, ByRef hKey As Long, ByVal pbKeyObject As Long, ByVal cbKeyObject As Long, ByVal pbSecret As Long, ByVal cbSecret As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptEncrypt Lib "bcrypt" (ByVal hKey As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal pPaddingInfo As Long, ByVal pbIV As Long, ByVal cbIV As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByRef cbResult As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptDecrypt Lib "bcrypt" (ByVal hKey As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal pPaddingInfo As Long, ByVal pbIV As Long, ByVal cbIV As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByRef cbResult As Long, ByVal dwFlags As Long) As Long

'CNG Buffers
Private bAuthData() As Byte
Private bAuthTag() As Byte
Private bIV() As Byte
Private bKey() As Byte
Private bReadIV() As Byte
Private bReadKey() As Byte
Private bWriteIV() As Byte
Private bWriteKey() As Byte
Private bInBuffer() As Byte
Private bOutBuffer() As Byte

Private bMsgType As Byte

'Counters
Private SEND_SEQ_NUM() As Byte
Private RECV_SEQ_NUM() As Byte

Private Type BCRYPT_KEY_LENGTHS_STRUCT
    dwMinLength As Long
    dwMaxLength As Long
    dwIncrement As Long
End Type

'Because VB6 does not support 64 bit (8 byte) long values, cbData has to be configured as
'a byte array. This structure is used repeatedly, and because the structure itself is not
'evenly devisible by 8, an additional long (lPad2) has been added to it.
Private Type BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO
    cbSize          As Long
    dwInfoVersion   As Long
    pbNonce         As Long
    cbNonce         As Long
    pbAuthData      As Long
    cbAuthData      As Long
    pbTag           As Long
    cbTag           As Long
    pbMacContext    As Long
    cbMacContext    As Long
    cbAAD           As Long
    lPad            As Long
    cbData(7)       As Byte
    dwFlags         As Long
    lPad2           As Long
End Type

Public Property Get AuthData() As Byte()
    AuthData = bAuthData
End Property

Public Property Let AuthData(bNewValue() As Byte)
    bAuthData = bNewValue
End Property

Public Property Get AuthTag() As Byte()
    AuthTag = bAuthTag
End Property

Public Property Let AuthTag(bNewValue() As Byte)
    bAuthTag = bNewValue
End Property

Public Function CryptData(sAlg As String, DeCrypt As Boolean) As Boolean
    Const Routine As String = "clsCrypt.CryptData"
    Dim hAlgorithm As Long 'Handle to algorithm sAlg"
    Dim hKey As Long 'Handle to Key
    Dim cbBlock As Long 'Encryption Block Size
    Dim bBuffer() As Byte 'Buffer for encryption object
    Dim authTagLengths As BCRYPT_KEY_LENGTHS_STRUCT 'Accepted lengths (12 to 16 step 1)
    Dim authInfo As BCRYPT_AUTHENTICATED_CIPHER_MODE_INFO 'Structure for Authenticated Encryption
    Dim pbOutput As Long 'Object length
    Dim cbResult As Long 'General use result
    Dim cbBytes As Long 'Input buffer size
    Dim pbTagLen As Long
    Dim dwFlags As Long 'Not used here - defaults to 0
    Dim Nonce() As Byte 'Nonce used in Galois/counter_mode (GCM)
    Dim SeqNum() As Byte
    Dim lRet As Long 'API call return value
    Dim N% 'Counter
    If DeCrypt Then
        bKey = bReadKey
        bIV = bReadIV
    Else
        bKey = bWriteKey
        bIV = bWriteIV
    End If
    ReDim Nonce(11) '12 byte Nonce set to all zeros
    'Recover handle to encryption algorithm
    lRet = BCryptOpenAlgorithmProvider(hAlgorithm, StrPtr(sAlg), StrPtr(MS_PRIMITIVE_PROVIDER), dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, COAP, Routine)
        GoTo ReleaseHandles
    End If
    'Get length of encryption object
    lRet = BCryptGetProperty(hAlgorithm, StrPtr(BCRYPT_OBJECT_LENGTH), VarPtr(pbOutput), 4, cbResult, dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    End If
    'Set the chaining mode (GCM)
    lRet = BCryptSetProperty(hAlgorithm, StrPtr(BCRYPT_CHAINING_MODE), StrPtr(BCRYPT_CHAIN_MODE_GCM), LenB(BCRYPT_CHAIN_MODE_GCM), dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGSK, Routine)
        GoTo ReleaseHandles
    End If
    'Recover Block length
    lRet = BCryptGetProperty(hAlgorithm, StrPtr(BCRYPT_BLOCK_LENGTH), VarPtr(cbBlock), 4, cbResult, dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    End If
    'Get the allowed Tag lengths (generally 12 bytes)
    lRet = BCryptGetProperty(hAlgorithm, StrPtr(BCRYPT_AUTH_TAG_LENGTH), VarPtr(authTagLengths), 12, cbResult, dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    Else
        pbTagLen = authTagLengths.dwMaxLength
        ReDim bAuthTag(pbTagLen - 1)
    End If
    ReDim bBuffer(pbOutput - 1) 'Clear the Object buffer
    'Generate the Object key information
    lRet = BCryptGenerateSymmetricKey(hAlgorithm, hKey, VarPtr(bBuffer(0)), GetbSize(bBuffer), VarPtr(bKey(0)), GetbSize(bKey), dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGSK, Routine)
        GoTo ReleaseHandles
    End If
    cbBytes = GetbSize(bInBuffer) 'Set the input buffer size
    If DeCrypt Then 'On decryption, demove AuthTag, but not record type
        cbBytes = cbBytes - pbTagLen
        CopyMemory bAuthTag(0), bInBuffer(cbBytes), pbTagLen 'Copy AuthTag from input
        ReDim Preserve bInBuffer(cbBytes - 1) 'Remove AuthTag, but not record type
    Else 'On encryption, add the Record Type
        ReDim Preserve bInBuffer(cbBytes)
        bInBuffer(cbBytes) = &H16
        cbBytes = cbBytes + 1
    End If
    If GetbSize(SEND_SEQ_NUM) < 8 Then ReDim SEND_SEQ_NUM(11) 'Initialize Sequence Number to all zeros
    If GetbSize(RECV_SEQ_NUM) < 8 Then ReDim RECV_SEQ_NUM(11) 'Initialize Sequence Number to all zeros
    If DeCrypt Then
        SeqNum = RECV_SEQ_NUM
    Else
        SeqNum = SEND_SEQ_NUM
    End If
    For N% = 0 To UBound(bIV) 'Create Nonce by combining the Handshake IV and Sequence Number
       Nonce(N%) = bIV(N%) Xor SeqNum(N%)
    Next N%
    'Get the encrypted length - should be same as Input length less AuthTag
    'Even though AES is considered a Block Cipher, in GCM mode it is considered a Stream cipher
    lRet = BCryptEncrypt(hKey, VarPtr(bInBuffer(0)), cbBytes, 0&, VarPtr(bIV(0)), cbBlock, 0&, 0, cbResult, dwFlags)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CE, Routine)
        GoTo ReleaseHandles
    End If
    ReDim bAuthData(4) 'Create Authdata using final record header
    bAuthData(0) = 23: bAuthData(1) = 3: bAuthData(2) = 3
    bAuthData(3) = Int(cbBytes / 256)
    bAuthData(4) = cbBytes Mod 256 + cbBlock
    'Populate authInfo structure
    authInfo.cbSize = Len(authInfo)
    authInfo.dwInfoVersion = 1
    authInfo.pbNonce = VarPtr(Nonce(0))
    authInfo.cbNonce = GetbSize(Nonce)
    authInfo.pbTag = VarPtr(bAuthTag(0))
    authInfo.cbTag = pbTagLen
    authInfo.pbAuthData = VarPtr(bAuthData(0))
    authInfo.cbAuthData = GetbSize(bAuthData)
    If DeCrypt Then
        ReDim bOutBuffer(cbResult - 1) 'Set the ouput buffer size
        lRet = BCryptDecrypt(hKey, VarPtr(bInBuffer(0)), cbBytes, VarPtr(authInfo), VarPtr(bIV(0)), 12, VarPtr(bOutBuffer(0)), cbResult, cbResult, dwFlags)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, CE, Routine)
            GoTo ReleaseHandles
        End If
        IncRecvSeqNum 'Advance the Receive Sequence Number
        ReDim Preserve bOutBuffer(cbResult - 2) 'Remove Record Type
    Else
        ReDim bOutBuffer(cbResult - 1) 'Set the ouput buffer size
        lRet = BCryptEncrypt(hKey, VarPtr(bInBuffer(0)), cbBytes, VarPtr(authInfo), VarPtr(bIV(0)), 12, VarPtr(bOutBuffer(0)), cbResult, cbResult, dwFlags)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, CE, Routine)
            GoTo ReleaseHandles
        End If
        IncSendSeqNum 'Advance the Send Sequence Number
        ReDim Preserve bOutBuffer(cbResult + cbBlock - 1) 'Add the AuthTag
        CopyMemory bOutBuffer(cbResult), ByVal VarPtr(bAuthTag(0)), cbBlock
    End If
    CryptData = True 'Success
ReleaseHandles:
    BCryptDestroyKey hKey
    BCryptCloseAlgorithmProvider hAlgorithm, 0
End Function


Public Sub IncRecvSeqNum(Optional flgClear As Boolean)
    Dim N%
    If flgClear Then
        ReDim RECV_SEQ_NUM(11)
        Exit Sub
    End If
    For N% = 11 To 7 Step -1
        If N% = 7 Then
            ReDim RECV_SEQ_NUM(11)
            Exit Sub
        End If
        If RECV_SEQ_NUM(N%) = 255 Then
            RECV_SEQ_NUM(N%) = 0
        Else
            RECV_SEQ_NUM(N%) = RECV_SEQ_NUM(N%) + 1
            Exit For
        End If
    Next N%
    'Debug.Print "Receive", RECV_SEQ_NUM(11)
End Sub

Public Sub IncSendSeqNum(Optional flgClear As Boolean)
    Dim N%
    If flgClear Then
        ReDim SEND_SEQ_NUM(11)
        Exit Sub
    End If
    For N% = 11 To 7 Step -1
        If N% = 7 Then
            ReDim SEND_SEQ_NUM(11)
            Exit Sub
        End If
        If SEND_SEQ_NUM(N%) = 255 Then
            SEND_SEQ_NUM(N%) = 0
        Else
            SEND_SEQ_NUM(N%) = SEND_SEQ_NUM(N%) + 1
            Exit For
        End If
    Next N%
    'Debug.Print "Send", SEND_SEQ_NUM(11)
End Sub

Public Property Get MsgType() As Byte
    MsgType = bMsgType
End Property

Public Property Let MsgType(bNewValue As Byte)
    bMsgType = bNewValue
End Property

Public Property Get InBuffer() As Byte()
    InBuffer = bInBuffer
End Property

Public Property Let InBuffer(bNewValue() As Byte)
    bInBuffer = bNewValue
End Property

Public Property Get OutBuffer() As Byte()
    OutBuffer = bOutBuffer
End Property

Public Property Let OutBuffer(bNewValue() As Byte)
    bOutBuffer = bNewValue
End Property

Public Property Let ReadIV(bNewValue() As Byte)
    bReadIV = bNewValue
End Property

Public Property Let ReadKey(bNewValue() As Byte)
    bReadKey = bNewValue
End Property

Public Property Let WriteIV(bNewValue() As Byte)
    bWriteIV = bNewValue
End Property

Public Property Let WriteKey(bNewValue() As Byte)
    bWriteKey = bNewValue
End Property

