VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "clsHKDF"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

'================================
'EVENTS
'================================
Public Event Error(ByVal Number As Long, Description As String, ByVal Source As String)

Private Const MS_PRIMITIVE_PROVIDER As String = "Microsoft Primitive Provider"

Private Type BCryptBuffer
    cbBuffer As Long                'Length of buffer, in bytes
    BufferType As Long              'Buffer type
    pvBuffer As Long                'Pointer to buffer
End Type

Private Type BCryptBufferDesc
    ulVersion As Long               'Version number
    cBuffers As Long                'Number of buffers
    pBuffers As Long                'Pointer to array of buffers
    rgBuffer(3) As BCryptBuffer     'Allowance for 4 buffers (more can be used)
End Type

'CNG API Declares
Private Declare Function BCryptOpenAlgorithmProvider Lib "bcrypt" (ByRef hAlgorithm As Long, ByVal pszAlgId As Long, ByVal pszImplementation As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptCloseAlgorithmProvider Lib "bcrypt" (ByVal hAlgorithm As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptGetProperty Lib "bcrypt" (ByVal hObject As Long, ByVal pszProperty As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByRef cbResult As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptCreateHash Lib "bcrypt" (ByVal hAlgorithm As Long, ByRef hHash As Long, ByVal pbHashObject As Long, ByVal cbHashObject As Long, ByVal pbSecret As Long, ByVal cbSecret As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptHashData Lib "bcrypt" (ByVal hHash As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptHash Lib "bcrypt.dll" (ByVal hAlgorithm As Long, ByVal pbSecret As Long, ByVal cbSecret As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal pbOutput As Long, ByVal cbOutput As Long) As Long
Private Declare Function BCryptFinishHash Lib "bcrypt" (ByVal hHash As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptDestroyHash Lib "bcrypt" (ByVal hHash As Long) As Long
Private Declare Function BCryptDuplicateHash Lib "bcrypt.dll" (ByVal hHash As Long, ByRef hNewHash As Long, ByVal pbHashObject As Long, ByVal cbHashObject As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptGenerateKeyPair Lib "bcrypt" (ByVal hAlgorithm As Long, ByRef hKey As Long, ByVal dwLength As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptFinalizeKeyPair Lib "bcrypt" (ByVal hKey As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptExportKey Lib "bcrypt.dll" (ByVal hKey As Long, ByVal hExportKey As Long, ByVal pszBlobType As Long, ByVal pbOutput As Long, ByVal cbOutput As Long, ByRef cbResult As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptImportKeyPair Lib "bcrypt.dll" (ByVal hAlgorithm As Long, ByVal hImportKey As Long, ByVal pszBlobType As Long, ByRef hKey As Long, ByVal pbInput As Long, ByVal cbInput As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptSecretAgreement Lib "bcrypt.dll" (ByVal hPrivKey As Long, ByVal hPubKey As Long, ByRef phSecret As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptDeriveKey Lib "bcrypt.dll" (ByVal hSharedSecret As Long, ByVal pwszKDF As Long, ByVal pParameterList As Long, ByVal pbDerivedKey As Long, ByVal cbDerivedKey As Long, ByRef pcbResult As Long, ByVal dwFlags As Long) As Long
Private Declare Function BCryptDestroyKey Lib "bcrypt" (ByVal hAesKey As Long) As Long
Private Declare Function BCryptDestroySecret Lib "bcrypt.dll" (ByVal hSecret As Long) As Long

'BCryptGetProperty strings (subset used here).
Private Const BCRYPT_OBJECT_LENGTH As String = "ObjectLength"
Private Const BCRYPT_HASH_LENGTH As String = "HashDigestLength"
Private Const BCRYPT_BLOCK_LENGTH As String = "BlockLength"

'Constants for Cryptography API error messages
'Private Const CCAP As String = "BCryptCloseAlgorithmProvider"
Private Const CCH As String = "BCryptCreateHash"
'Private Const CD As String = "BCryptDecrypt"
'Private Const CDH As String = "BCryptDestroyHash"
'Private Const CDK As String = "BCryptDestroyKey"
'Private Const CDRK As String = "BCryptDerivrKey"
'Private Const CE As String = "BCryptEncrypt"
'Private Const CEA As String = "BCryptEnumAlgorithms"
'Private Const CEK As String = "BCryptExportKey"
'Private Const CFKP As String = "BCryptFinalizeKeyPair"
Private Const CFH As String = "BCryptFinishHash"
'Private Const CGKP As String = "BCryptGenerateKeyPair"
Private Const CGP As String = "BCryptGetProperty"
'Private Const CGR As String = "BCryptGenRandom"
'Private Const CGSK As String = "BCryptGenerateSymmetricKey"
Private Const CHD As String = "BCryptHashData"
'Private Const CIKP As String = "BCryptImportKeyPair"
Private Const COAP As String = "BCryptOpenAlgorithmProvider"
'Private Const CSA As String = "BCryptSecretAgreement"
'Private Const CSH As String = "BCryptSignHash"
'Private Const CSP As String = "BCryptSetProperty"
'Private Const CVS As String = "BCryptVerifySignature"
Private Const CDuH As String = "BCryptDuplicateHash"

Private bInBuffer() As Byte
Private bOutBuffer() As Byte

Public Function AddToHash(sAlg As String, DataToHash() As Byte, hHash As Long, lHash As Long, bHashObject() As Byte) As Long
    Const Routine As String = "clsCNG.AddToHash"
    Dim hAlg As Long
    Dim cbHashObject As Long
    Dim cbResult As Long
    Dim bHash() As Byte
    Dim lRet As Long
    If hHash = 0 Then
        lRet = BCryptOpenAlgorithmProvider(hAlg, StrPtr(sAlg), StrPtr(MS_PRIMITIVE_PROVIDER), 0)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, COAP, Routine)
            GoTo ReleaseHandles
        End If
        lRet = BCryptGetProperty(hAlg, StrPtr("ObjectLength"), VarPtr(cbHashObject), 4, cbResult, 0)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, CGP, Routine)
            GoTo ReleaseHandles
        End If
        lRet = BCryptGetProperty(hAlg, StrPtr("HashDigestLength"), VarPtr(lHash), 4, cbResult, 0)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, CGP, Routine)
            GoTo ReleaseHandles
        End If
        ReDim bHashObject(cbHashObject - 1)
        lRet = BCryptCreateHash(hAlg, hHash, VarPtr(bHashObject(0)), cbHashObject, 0&, 0, 0)
        If lRet <> 0 Then
            RaiseEvent Error(lRet, CCH, Routine)
            GoTo ReleaseHandles
        End If
    End If
    lRet = BCryptHashData(hHash, VarPtr(DataToHash(0)), GetbSize(DataToHash), 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CHD, Routine)
        GoTo ReleaseHandles
    End If
    AddToHash = hHash
ReleaseHandles:
    BCryptCloseAlgorithmProvider hAlg, 0
End Function

Public Function CopyHash(hHashToCopy As Long, HashObject() As Byte) As Long
    Const Routine As String = "clsCNG.CopyHash"
    Dim lRet As Long
    Dim hNewHash As Long
    lRet = BCryptDuplicateHash(hHashToCopy, hNewHash, VarPtr(HashObject(0)), GetbSize(HashObject), 0&)
    If lRet = 0 Then
        CopyHash = hNewHash
    Else
        RaiseEvent Error(lRet, CDuH, Routine)
        CopyHash = 0
    End If
End Function

Public Function CreateInfo(sLabel As String, bHash() As Byte, ExpandLen As Long) As Byte()
    Dim bInfo() As Byte
    Dim bLabel() As Byte
    Dim LabelSize As Long
    Dim HashSize As Long
    LabelSize = Len(sLabel)
    bLabel = StrToByte(sLabel)
    HashSize = GetbSize(bHash)
    If HashSize = 1 Then HashSize = 0
    ReDim bInfo(LabelSize + HashSize + 3)
    CopyMemory bInfo(3), bLabel(0), LabelSize
    If HashSize > 0 Then
        CopyMemory bInfo(LabelSize + 4), bHash(0), HashSize
    End If
    bInfo(1) = ExpandLen
    bInfo(2) = LabelSize
    bInfo(LabelSize + 3) = HashSize
    CreateInfo = bInfo
End Function

Public Function Extract(HashAlg As String, bSalt() As Byte, bIKM() As Byte) As Boolean
    bInBuffer = bIKM
    If GetbSize(bSalt) = 0 Then ReDim bSalt(0)
    Call HMAC(HashAlg, VarPtr(bSalt(0)), GetbSize(bSalt))
End Function

Public Function GetECCKey(ByVal pAlg As Long, ByVal KeyLen As Long, bPublicECCKey() As Byte, bPrivateECCKey() As Byte) As Byte()
    Dim lRet As Long
    Dim hProv As Long
    Dim hPrivKey As Long
    Dim cbPubBlob As Long
    Dim cbPriBlob As Long
    Dim cbPubBlobS As Long
    Dim bPubBlob() As Byte
    Dim bPriBlob() As Byte
    Dim hPubKey As Long
    Dim hAgreedSecret As Long
    Dim cbAgreedSecret As Long
    Dim bAgreedSecret() As Byte
    Dim AuxFlg As Boolean
    Dim lPntr As Long
    Dim ParameterList As BCryptBufferDesc
    lRet = BCryptOpenAlgorithmProvider(hProv, pAlg, StrPtr("Microsoft Primitive Provider"), 0)
    If lRet <> 0 Then
        ReDim bAgreedSecret(0)
        bAgreedSecret(0) = 1
        GoTo ReleaseHandles
    End If
    cbPubBlob = GetbSize(bPublicECCKey)
    cbPriBlob = GetbSize(bPrivateECCKey)
    If cbPubBlob < 2 And cbPriBlob < 2 Then 'Return internal Public/Private Keys
        lRet = BCryptGenerateKeyPair(hProv, hPrivKey, KeyLen, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 2
            GoTo ReleaseHandles
        End If
        lRet = BCryptFinalizeKeyPair(hPrivKey, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 3
            GoTo ReleaseHandles
        End If
        'Export private key
        lRet = BCryptExportKey(hPrivKey, 0&, StrPtr("ECCPRIVATEBLOB"), 0&, 0&, cbPriBlob, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 4
            GoTo ReleaseHandles
        End If
        ReDim bPriBlob(cbPriBlob - 1)
        lRet = BCryptExportKey(hPrivKey, 0&, StrPtr("ECCPRIVATEBLOB"), VarPtr(bPriBlob(0)), cbPriBlob, cbPriBlob, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 5
            GoTo ReleaseHandles
        End If
        'Export public key
        lRet = BCryptExportKey(hPrivKey, 0&, StrPtr("ECCPUBLICBLOB"), 0&, 0&, cbPubBlob, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 6
            GoTo ReleaseHandles
        End If
        ReDim bPubBlob(cbPubBlob - 1)
        lRet = BCryptExportKey(hPrivKey, 0&, StrPtr("ECCPUBLICBLOB"), VarPtr(bPubBlob(0)), cbPubBlob, cbPubBlob, 0)
        If lRet <> 0 Then
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 7
            GoTo ReleaseHandles
        End If
        bPublicECCKey = bPubBlob
        bPrivateECCKey = bPriBlob
        ReDim bAgreedSecret(0)
        bAgreedSecret(0) = 0 '0 indicates success
        GoTo ReleaseHandles
    End If
    cbPubBlobS = GetbSize(bPrivateECCKey)
    lRet = BCryptImportKeyPair(hProv, 0&, StrPtr("ECCPRIVATEBLOB"), hPrivKey, VarPtr(bPrivateECCKey(0)), cbPubBlobS, 0)
    If lRet <> 0 Then
        ReDim bAgreedSecret(0)
        bAgreedSecret(0) = 8
        GoTo ReleaseHandles
    End If
    'Import Server's public key
    cbPubBlobS = GetbSize(bPublicECCKey)
    lRet = BCryptImportKeyPair(hProv, 0&, StrPtr("ECCPUBLICBLOB"), hPubKey, VarPtr(bPublicECCKey(0)), cbPubBlobS, 0)
    If lRet <> 0 Then
        ReDim bAgreedSecret(0)
        bAgreedSecret(0) = 9
        GoTo ReleaseHandles
    End If
    'Generate the agreed secret
    lRet = BCryptSecretAgreement(hPrivKey, hPubKey, hAgreedSecret, 0)
    If lRet <> 0 Then
        ReDim bAgreedSecret(0)
        bAgreedSecret(0) = 10
        GoTo ReleaseHandles
    End If
    'Get Agreed Secret length
    lRet = BCryptDeriveKey(hAgreedSecret, StrPtr("TRUNCATE"), 0, 0&, 0, cbAgreedSecret, 0)
    If lRet = 0 Then
        ReDim bAgreedSecret(cbAgreedSecret - 1)
    Else
        lPntr = PeekW(hAgreedSecret + 12) + 12
        CopyMemory ByVal VarPtr(cbAgreedSecret), ByVal lPntr, 4
        If cbAgreedSecret > 0 Then
            AuxFlg = True
            ReDim bAgreedSecret(cbAgreedSecret - 1)
        Else
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 11
            GoTo ReleaseHandles
        End If
    End If
    'Get Agreed Secret
    lRet = BCryptDeriveKey(hAgreedSecret, StrPtr("TRUNCATE"), 0, VarPtr(bAgreedSecret(0)), cbAgreedSecret, cbAgreedSecret, 0)
    If lRet <> 0 Then
        If AuxFlg Then
            CopyMemory bAgreedSecret(0), ByVal lPntr + 4, cbAgreedSecret
        Else
            ReDim bAgreedSecret(0)
            bAgreedSecret(0) = 12
            GoTo ReleaseHandles
        End If
    End If
   'build KDF parameter list
'    With ParameterList
'        .rgBuffer(0).BufferType = 0&
'        .rgBuffer(0).cbBuffer = LenB("SHA256") + 2
'        .rgBuffer(0).pvBuffer = StrPtr("SHA256")
'        .cBuffers = 1
'        .pBuffers = VarPtr(.rgBuffer(0))
'        .ulVersion = 0 'BCRYPTBUFFER_VERSION
'    End With
'    'derive keys from secret using specified KDF
'    lRet = BCryptDeriveKey(hAgreedSecret, StrPtr("HASH"), VarPtr(ParameterList), 0&, 0, cbAgreedSecret, 0)
'    If lRet <> 0 Then
'        ReDim bAgreedSecret(0)
'        bAgreedSecret(0) = 11
'        GoTo ReleaseHandles
'    End If
'    ReDim bAgreedSecret(cbAgreedSecret - 1)
'    lRet = BCryptDeriveKey(hAgreedSecret, StrPtr("HASH"), VarPtr(ParameterList), VarPtr(bAgreedSecret(0)), cbAgreedSecret, cbAgreedSecret, 0)
'    If lRet <> 0 Then
'        ReDim bAgreedSecret(0)
'        bAgreedSecret(0) = 12
'        GoTo ReleaseHandles
'    End If
ReleaseHandles:
    GetECCKey = bAgreedSecret
    If (hPubKey) Then BCryptDestroyKey (hPubKey)
    If (hPrivKey) Then BCryptDestroyKey (hPrivKey)
    If (hProv) Then BCryptCloseAlgorithmProvider hProv, 0
    If (hAgreedSecret) Then BCryptDestroySecret hAgreedSecret
End Function

Public Function GetHash(hHash As Long, lHash As Long) As Byte()
    Const Routine As String = "clsCNG.GetHash"
    Dim bHash() As Byte
    Dim lRet As Long
    ReDim bHash(lHash - 1)
    lRet = BCryptFinishHash(hHash, VarPtr(bHash(0)), lHash, 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CFH, Routine)
        GoTo ReleaseHandles
    End If
    GetHash = bHash
ReleaseHandles:
    BCryptDestroyHash hHash: hHash = 0
End Function

Public Function HashData(HashAlg As String, bInBuffer() As Byte) As Byte()
    Dim hAlgorithm As Long
    Dim cbResult As Long
    Dim cbHashObject As Long
    Dim cbHashBlock As Long
    Dim HashObject() As Byte
    Dim Hash() As Byte
    Dim hHash As Long
    If BCryptOpenAlgorithmProvider(hAlgorithm, StrPtr(HashAlg), StrPtr("Microsoft Primitive Provider"), 0) Then GoTo ReleaseHandles
    If BCryptGetProperty(hAlgorithm, StrPtr("HashDigestLength"), VarPtr(cbHashBlock), 4, cbResult, 0) Then GoTo ReleaseHandles
    ReDim Hash(cbHashBlock - 1)
    On Error GoTo NotW10
    If BCryptHash(hAlgorithm, 0&, 0, VarPtr(bInBuffer(0)), GetbSize(bInBuffer), VarPtr(Hash(0)), cbHashBlock) Then GoTo ReleaseHandles
    HashData = Hash
    GoTo ReleaseHandles
NotW10:
    On Error GoTo 0
    If BCryptGetProperty(hAlgorithm, StrPtr("ObjectLength"), VarPtr(cbHashObject), 4, cbResult, 0) Then GoTo ReleaseHandles
    ReDim HashObject(cbHashObject - 1)
    If BCryptCreateHash(hAlgorithm, hHash, VarPtr(HashObject(0)), cbHashObject, 0&, 0, 0) Then GoTo ReleaseHandles
    Hash = bInBuffer
    If GetbSize(Hash) = 0 Then ReDim Hash(0) 'Allows zero length data
    If BCryptHashData(hHash, VarPtr(Hash(0)), GetbSize(bInBuffer), 0) Then GoTo ReleaseHandles
    ReDim Hash(cbHashBlock - 1)
    If BCryptFinishHash(hHash, VarPtr(Hash(0)), cbHashBlock, 0) Then GoTo ReleaseHandles
    HashData = Hash
ReleaseHandles:
    BCryptDestroyHash hHash
    BCryptCloseAlgorithmProvider hAlgorithm, 0
End Function

Public Function Expand(HashAlg As String, bPRK() As Byte, bInfo() As Byte, ExpandLen As Long) As Boolean
    Dim bTmp() As Byte 'Temporary working input to HMAC
    Dim bResult() As Byte 'Result of individual HMAC
    Dim ResultLen As Long
    Dim bFinal() As Byte 'Final HMAC
    Dim FinalLen As Long 'Length of ongoing Expand
    Dim InfoLen As Long
    Dim LoopCount As Long
    LoopCount = 1
    InfoLen = GetbSize(bInfo)
    Do Until FinalLen > ExpandLen
        ReDim bTmp(ResultLen + InfoLen)
        'First loop uses zero length bResult
        If ResultLen > 0 Then CopyMemory bTmp(0), bResult(0), ResultLen
        If InfoLen > 0 Then CopyMemory bTmp(ResultLen), bInfo(0), InfoLen
        bTmp(GetbSize(bTmp) - 1) = LoopCount
        bInBuffer = bTmp
        Call HMAC(HashAlg, VarPtr(bPRK(0)), GetbSize(bPRK))
        bResult = bOutBuffer
        ResultLen = GetbSize(bResult)
        FinalLen = FinalLen + ResultLen
        Call AddByte(bFinal, bOutBuffer, ResultLen)
        LoopCount = LoopCount + 1
    Loop
    ReDim Preserve bFinal(ExpandLen - 1)
    bOutBuffer = bFinal
End Function


Public Function HMAC(sAlgID As String, hKey As Long, KeyLen As Long) As Boolean
    Const Routine As String = "clsHKDF.HMAC"
    Const BCRYPT_ALG_HANDLE_HMAC_FLAG = &H8
    Dim hAlgorithm As Long
    Dim cbHashBlock As Long
    Dim cbHashObject As Long
    Dim cbResult As Long
    Dim hHash As Long
    Dim Hash() As Byte
    Dim lRet As Long
    If GetbSize(bInBuffer) = 0 Then Exit Function
    lRet = BCryptOpenAlgorithmProvider(hAlgorithm, StrPtr(sAlgID), StrPtr(MS_PRIMITIVE_PROVIDER), BCRYPT_ALG_HANDLE_HMAC_FLAG)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, COAP, Routine)
        GoTo ReleaseHandles
    End If
    lRet = BCryptGetProperty(hAlgorithm, StrPtr(BCRYPT_HASH_LENGTH), VarPtr(cbHashBlock), 4, cbResult, 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    End If
    On Error GoTo NotW10
    'If not Win 10, next function produces Run-time error '453':
    'Can't find DLL entry point BCryptHash in bcrypt.dll
    lRet = BCryptHash(hAlgorithm, 0&, 0, VarPtr(bInBuffer(0)), GetbSize(bInBuffer), VarPtr(Hash(0)), cbHashBlock)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    End If
    bOutBuffer = Hash
    GoTo ReleaseHandles 'Otherwise we are done
NotW10:
    On Error GoTo 0
    lRet = BCryptGetProperty(hAlgorithm, StrPtr(BCRYPT_OBJECT_LENGTH), VarPtr(cbHashObject), 4, cbResult, 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CGP, Routine)
        GoTo ReleaseHandles
    End If
    ReDim HashObject(cbHashObject - 1)
    lRet = BCryptCreateHash(hAlgorithm, hHash, VarPtr(HashObject(0)), cbHashObject, hKey, KeyLen, 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CCH, Routine)
        GoTo ReleaseHandles
    End If
    Hash = bInBuffer
    lRet = BCryptHashData(hHash, VarPtr(Hash(0)), GetbSize(bInBuffer), 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CHD, Routine)
        GoTo ReleaseHandles
    End If
    ReDim Hash(cbHashBlock - 1)
    lRet = BCryptFinishHash(hHash, VarPtr(Hash(0)), cbHashBlock, 0)
    If lRet <> 0 Then
        RaiseEvent Error(lRet, CFH, Routine)
        GoTo ReleaseHandles
    End If
    bOutBuffer = Hash
    HMAC = True
ReleaseHandles:
    BCryptDestroyHash hHash
    BCryptCloseAlgorithmProvider hAlgorithm, 0
End Function

Public Property Get InBuffer() As Byte()
    InBuffer = bInBuffer
End Property

Public Property Let InBuffer(NewValue() As Byte)
    bInBuffer = NewValue
End Property

Public Property Get OutBuffer() As Byte()
    OutBuffer = bOutBuffer
End Property

Public Property Let OutBuffer(NewValue() As Byte)
    bOutBuffer = NewValue
End Property

Private Function PeekW(ByVal lpdwData As Long) As Long
    CopyMemory PeekW, ByVal lpdwData, 4
End Function

